<?php

namespace Webt\Services;

defined( '_JEXEC' ) or die;

/**
 * It provides functions that help extracting translatable content from HTML; restore it in original from.
 *
 * @license     GNU General Public License version 2 or later, see License.txt
 */
class TranslationHelper {
	/**
	 * List of translatable HTML attributes.
	 *
	 * @var array
	 */
	private static $translatableHtmlAttributes = array( 'alt' );
	/**
	 * Matches placeholders for Joomla plugins (e.g. {loadmodule ...}, {field ...}, {gallery}/images{/gallery}) that can be part of article content.
	 * Also matches any whitespaces around the placeholders.
	 *
	 * @var string
	 */
	private static $nonTranslatableStringPattern = '/\s*\{([a-zA-Z]+)(?:\s+[^{}]*)?\}(?:[^{}]*\{\/\1\})?\s*/';

	/**
	 * Converts strings to HTML and extracts translatable text and attribute values
	 *
	 * @param string $values
	 *   Translatable string values, that can be HTML.
	 *
	 * @return string[]
	 *   Translatable strings
	 */
	public static function extractTranslatableValues( $values ) {
		$texts_to_translate = array();
		foreach ( $values as $v ) {
			$html_string        = self::wrapAsHtml( $v );
			$texts              = self::extractTextsFromHtml( $html_string );
			$texts_to_translate = array_merge( $texts_to_translate, $texts );
		}
		return $texts_to_translate;
	}

	/**
	 * Merges raw translation list to a list of translations for original strings
	 *
	 * @param string[] $original_strings Original values.
	 * @param string[] $extracted_values Extracted HTML translatable values.
	 * @param string[] $raw_translations Raw translations from MT service.
	 * @return string[] Translations for original values.
	 */
	public static function mergeTranslatedValues( $original_strings, $extracted_values, $raw_translations ) {
		$translations = self::restoreSpaces( $extracted_values, $raw_translations );
		$translations = self::replaceStringTranslations( $original_strings, $translations );
		return $translations;
	}

	/**
	 * Goes through HTML DOM and extracts translatable text content and attribute values.
	 *
	 * @param string $html_content HTML string.
	 * @return string[] Translatable strings.
	 */
	private static function extractTextsFromHtml( $html_content ) {
		$texts_to_translate = array();
		$dom                = self::loadHTML( $html_content );
		// Extract text content from all text nodes.
		self::traverseDom(
			$dom,
			function( $node, $attr = null ) use ( &$texts_to_translate ) {
				if ( $attr ) {
					$texts_to_translate[] = $node->getAttribute( $attr );
				} else {
					$translatable_parts = self::splitNodeTranslatableParts( $node->nodeValue )[0];
					$texts_to_translate = array_merge( $texts_to_translate, $translatable_parts );
				}
			}
		);
		return $texts_to_translate;
	}

	/**
	 * Inserts translated node & attribute values into original strings.
	 *
	 * @param string[] $source_values
	 *   Original strings that were translated.
	 * @param string[] $translated_values
	 *   String value translations.
	 *
	 * @return string[]
	 *   Translated strings
	 */
	private static function replaceStringTranslations( $original_values, $translated_values ) {
		$index        = 0;
		$translations = array();
		foreach ( $original_values as $v ) {
			$html_content = self::wrapAsHtml( $v );
			$dom          = self::loadHTML( $html_content );

			self::traverseDom(
				$dom,
				function( $node, $attr = null ) use ( &$translated_values, &$index ) {
					if ( $attr ) {
						$node->setAttribute( $attr, $translated_values[ $index++ ] );
					} else {
						$parts              = self::splitNodeTranslatableParts( $node->nodeValue );
						$translatable_count = count( $parts[0] );
						for ( $i = 0; $i < $translatable_count; $i++ ) {
							$parts[0][ $i ] = $translated_values[ $index++ ]; // replace translatable parts with their translations.
						}
						$node->nodeValue = self::nodePartsToString( $parts );
					}
				}
			);

			$wrapper        = $dom->documentElement->lastChild->lastChild;
			$translation    = self::getInnerHtml( $wrapper );
			$translations[] = $translation;
		}
		return $translations;
	}

	/**
	 * Traverses each HTML element in DOM and executes callback when text or attribute value is found.
	 *
	 * @param DOMDocument $dom HTML DOM.
	 * @param callable    $callback Callback function.
	 * @return void
	 */
	private static function traverseDom( $dom, $callback ) {
		$text_nodes = $dom->getElementsByTagName( '*' );
		foreach ( $text_nodes as $node ) {
			foreach ( $node->childNodes as $child ) {
				if ( $child->nodeType === XML_TEXT_NODE && trim( $child->nodeValue ) !== '' ) {
					$callback( $child );
				}
			}
			foreach ( self::$translatableHtmlAttributes as $attr ) {
				if ( $node->hasAttribute( $attr ) ) {
					$callback( $node, $attr );
				}
			}
		}
	}

	/**
	 * Restores spaces around HTML tags removed by MT service.
	 *
	 * @param string[] $original_values Original values.
	 * @param string[] $translated_values Translated values.
	 *
	 * @return string[]
	 *   Translations
	 */
	private static function restoreSpaces( $original_values, $translated_values ) {
		// Fix leading, trailing & double spaces.
		$count = count( $original_values );
		for ( $i = 0; $i < $count; $i++ ) {
			// Replace different types of UTF-8 spaces (incl. nbsp) with regular space.
			$original    = preg_replace( '%\s+%u', ' ', $original_values[ $i ] );
			$translation = preg_replace( '%\s+%u', ' ', $translated_values[ $i ] );

			if ( ctype_space( substr( $original, 0, 1 ) ) && ! ctype_space( substr( $translation, 0, 1 ) ) ) {
				$translated_values[ $i ] = substr( $original, 0, 1 ) . $translated_values[ $i ];
			}
			if ( ctype_space( substr( $original, -1 ) ) && ! ctype_space( substr( $translation, -1 ) ) ) {
				$translated_values[ $i ] = $translated_values[ $i ] . substr( $original, -1 );
			}
			$translated_values[ $i ] = str_replace( '  ', ' ', $translated_values[ $i ] );
		}
		return $translated_values;
	}

	/**
	 * Loads HTML content into a DOMDocument object.
	 *
	 * @param string $html The HTML content to be loaded.
	 * @return \DOMDocument The DOMDocument object representing the loaded HTML content.
	 */
	private static function loadHtml( $html ) {
		$document = <<<EOD
		<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
		<html xmlns="http://www.w3.org/1999/xhtml">
		<head><meta http-equiv="Content-Type" content="text/html; charset=utf-8" /></head>
		<body>!html</body>
		</html>
		EOD;
		// PHP's \DOMDocument::saveXML() encodes carriage returns as &#13; so normalize all newlines to line feeds.
		$html = str_replace(
			array(
				"\r\n",
				"\r",
			),
			"\n",
			$html
		);
		// PHP's \DOMDocument serialization adds extra whitespace when the markup
		// of the wrapping document contains newlines, so ensure we remove all
		// newlines before injecting the actual HTML body to be processed.
		$document = strtr(
			$document,
			array(
				"\n"    => '',
				'!html' => $html,
			)
		);
		$dom      = new \DOMDocument();
		// Ignore warnings during HTML soup loading.
		@$dom->loadHTML( $document, LIBXML_NOBLANKS );
		return $dom;
	}

	/**
	 * Get inner HTML content of a node.
	 *
	 * @param \DOMNode $node HTML node.
	 * @return string HTML content as a string.
	 */
	private static function getInnerHtml( $node ) {
		$innerHTML = '';
		$children  = $node->childNodes;
		foreach ( $children as $child ) {
			$innerHTML .= $child->ownerDocument->saveXML( $child );
		}
		return $innerHTML;
	}

	/**
	 * Wrap translatable string in HTML tags.
	 *
	 * @param string $string
	 *   Source string.
	 *
	 * @return string
	 *   HTML node string
	 */
	private static function wrapAsHtml( $string ) {
		return '<div>' . $string . '</div>';
	}

	/**
	 * Splits translatable and non-translatable parts of a node.
	 *
	 * @param string $string Source string.
	 * @return array Strings of node grouped by translatable/non-translatable status.
	 */
	private static function splitNodeTranslatableParts( $string ) {
		preg_match_all( self::$nonTranslatableStringPattern, $string, $matches, PREG_OFFSET_CAPTURE );
		$matches = $matches[0];

		$translatable_parts     = array();
		$non_translatable_parts = array( '' ); // empty value helps to restore string afterwards by ensuring it always starts with non-translatable part.

		if ( empty( $matches ) ) {
			$translatable_parts = array( $string );
		} else {
			$previous_index = 0;

			foreach ( $matches as $match ) {
				$match_position = $match[1];
				if ( $match_position === 0 ) {
					$non_translatable_parts[0] = $match[0]; // overwrite empty value if string actually starts with non-translatable part.
				} else {
					$non_translatable_parts[] = $match[0];
				}
				if ( $match_position > $previous_index ) {
					$translatable_parts[] = substr( $string, $previous_index, $match_position - $previous_index );
				}
				$previous_index = $match_position + strlen( $match[0] );
			}

			if ( $previous_index < strlen( $string ) ) {
				$translatable_parts[] = substr( $string, $previous_index );
			}
		}
		return array( $translatable_parts, $non_translatable_parts );
	}

	/**
	 * Combines node translatable and non translatable parts back into string
	 *
	 * @param array $parts Array of translatable and non-translatable string arrays.
	 * @return string Node value.
	 */
	private static function nodePartsToString( $parts ) {
		$combined               = array();
		$translatable_parts     = $parts[0];
		$non_translatable_parts = $parts[1];
		$count                  = max( count( $translatable_parts ), count( $non_translatable_parts ) );
		for ( $i = 0; $i < $count; $i++ ) {
			if ( isset( $non_translatable_parts[ $i ] ) ) {
				$combined[] = $non_translatable_parts[ $i ];
			}
			if ( isset( $translatable_parts[ $i ] ) ) {
				$combined[] = $translatable_parts[ $i ];
			}
		}
		return implode( '', $combined );
	}
}
